/*
 * SOCKLAB - "laboratoire" d'etude des sockets internet - ENSIMAG - Juillet 93
 *
 * prim.c: commandes associees a chaque primitive des sockets
 *
 */
/*Modifie par P. SIcard 09 97 passage sur PC : probleme de little./big endian */

/* cc96-1 : remplacement de bzero par memset */
/* cc96-2 : value type used inconsistently
    getpid           	llib-lc:unistd.h(176) long () :: prim.c(51) int ()
    strlen           	llib-lc:string.h(59) unsigned int () :: prim.c(386) int ()
    strlen           	llib-lc:string.h(59) unsigned int () :: prim.c(459) int ()
    strlen           	llib-lc:string.h(59) unsigned int () :: prim.c(548) int ()
*/

#include "socklab.h"
#include <stdlib.h>

char* versionPrim = "prim.c : $Revision: 71 $ du $Date: 2007-05-03 15:35:56 +0200 (Thu, 03 May 2007) $ par $Author: rousseau $";
 
/* Toutes les primitives n'ont que deux parametres: argc et argv qui ont le
 * meme sens que pour main(). Ici, ils s'appliquent aux arguments qui figuraient
 * sur la ligne de commandes saisie par l'utilisateur.
 */

/* Primitive socket()
 *=======================================================================
 *
 */

int socket_call(argc, argv)
	int	argc;
	char	*argv[];
{
	int             s;
	int             proto;
	static t_item   list[] = {{"tcp", SOCK_STREAM}, {"udp", SOCK_DGRAM}};
	if (nbsock == MAXSOCK)
	{
		printf ("La table interne des sockets est pleine.\n");
		return (-1);
	}

	if (argc>1)
		get_choice("Protocole", list, 2, argv[1],  &proto);
	else
		get_choice("Protocole", list, 2, "", &proto);

/* creation de la socket */
	s = socket(AF_INET, proto, 0);
	if (s<0)
	{
		ERREUR("socket()");
		return(-1);
	}

/* pour la reception de SIGIO */
	if (fcntl(s, F_SETOWN, getpid()) <0 )
	{
		ERREUR("fcntl()-SET_OWN");
		if (close(s)==-1)
			ERREUR("close()");
		
		return(-1);
	}

	if (DFT_MODE)
		printf("La socket est identifiee par l'identificateur %d\n", s);

/* modification de la table */
	sock[nbsock] = s;
	dft_sock = nbsock++;
	return(dft_sock);
}


/* Primitive bind()
 *=======================================================================
 *
 */

int bind_call(argc, argv)
	int	argc;
	char	*argv[];
{
	struct sockaddr_in sa;
	socklen_t       lensa = sizeof (struct sockaddr_in);
	int             port = -1;
	unsigned long   addr = INADDR_NONE;
	int             so;


/* id socket ? */
	if (argc>1)
		get_id_sock(argv[1], &so);
	else
		get_id_sock("", &so);

/* host ? */
	if (argc>2)
		get_host(argv[2], &addr);
	else
		get_host("", &addr);

/* port ? */
	if (argc>3)
		get_port(argv[3], &port);
	else
		get_port("", &port);

/* Initialisation de l'adresse */

	/* bzero((char *)&sa, sizeof(struct sockaddr_in)); cc96-1 */
	memset ((char *) &sa, 0, sizeof (struct sockaddr_in));

	sa.sin_family = AF_INET;
	port = htons ((u_short) port); /* Modif P. Sicard */
	sa.sin_port =   port;
	sa.sin_addr.s_addr = addr; /* attention pas de htonl ici ??? */

/* attachement de la socket */
	if (bind (sock[so], (struct sockaddr *) & sa, sizeof (sa)) < 0)
	{
		ERREUR ("bind()");
		return (-1);
	}

	if (port == 0)		/* obtention du port choisi par le systeme */
	{
		getsockname (sock[so], (struct sockaddr *) & sa, &lensa);
		port = ntohs (sa.sin_port); /* Modif P. Sicard */


		if (DFT_MODE)
			printf ("Le port %d a ete affecte a la socket\n",
				port); /* Modif P. Sicard */

	}

	return(port);
}


/* Primitive listen()
 *=======================================================================
 *
 */

int listen_call(argc, argv)
	int	argc;
	char	*argv[];
{
	int	nb = -1;
	int	so;


/* id socket ? */
	if (argc>1)
		get_id_sock(argv[1], &so);
	else
		get_id_sock("", &so);

/* backlog ? */
	nb = 5;
	if (argc>2)
		get_nb("Nb de requetes", argv[2], &nb);
	else
		get_nb("Nb de requetes", "", &nb);

/* mise en ecoute */
if (listen(sock[so], nb)<0)
	{
		ERREUR("listen()");
		return(-1);
	}

	return(0);
}


/* Primitive accept()
 *=======================================================================
 *
 */

int accept_call(argc, argv)
	int	argc;
	char	*argv[];
{
	struct sockaddr_in sa;
	socklen_t lensa = sizeof(struct sockaddr_in);
	struct hostent *hp;
	int	newso;
	int	so;


	if (nbsock==MAXSOCK)
	{
		printf("La table interne des sockets est pleine.\n");
		return(-1);
	}

/* id socket ? */
	if (argc>1)
		get_id_sock(argv[1], &so);
	else
		get_id_sock("", &so);

	newso = accept(sock[so], (struct sockaddr *)&sa, &lensa);
	if (newso<0)
	{
		ERREUR("accept()");
		return(-1);
	}

/* identification de l'appel entrant */
	hp = gethostbyaddr( (char *)&sa.sin_addr, sizeof(sa.sin_addr), AF_INET);

	if (hp == NULL)
		printf ("Un appel de %s (%d) a ete intercepte.\n",
			inet_ntoa (sa.sin_addr), htons(sa.sin_port));
			/*modif P Sicard 2001 ntohs*/
	else
		printf ("Un appel de %s (%d) a ete intercepte.\n",
			hp->h_name, ntohs(sa.sin_port));
			/*modif P Sicard 2001 ntohs*/

	printf("La connexion est etablie sous l'identificateur %d.\n", newso);

/* modification de la table */
	sock[nbsock] = newso;
	dft_sock = nbsock++;
	return(dft_sock);
}


/* Primitive connect()
 *=======================================================================
 *
 */

int connect_call(argc, argv)
	int	argc;
	char	*argv[];
{
	int	port = -1;
	u_long	addr = INADDR_NONE;
	struct sockaddr_in	sa;
	int	so;


/* id socket ? */
	if (argc>1)
		get_id_sock(argv[1], &so);
	else
		get_id_sock("", &so);

/* host ? */
	if (argc>2)
		get_host(argv[2], &addr);
	else
		get_host("", &addr);

/* port ? */
	if (argc>3)
		get_port(argv[3], &port);
	else
		get_port("", &port);

/* initialisation de l'adresse */

	/* bzero((char *)&sa, sizeof(struct sockaddr_in)); cc96-1 */
	memset ((char *) &sa, 0, sizeof (struct sockaddr_in));

	sa.sin_family = AF_INET;
	sa.sin_port = htons ((u_short) port);
	sa.sin_addr.s_addr = addr;  /* pas de htonl ?? Modif P. Sicard */

/* connexion */
	if (connect(sock[so], (struct sockaddr *)&sa, sizeof(sa))<0)
	{
		ERREUR("connect()");
		return(-1);
	}

	printf("Connexion etablie.\n");

/* mise a jour de la table */
	return(0);
}


/* Primitive close()
 *=======================================================================
 *
 */

int close_call(argc, argv)
	int	argc;
	char	*argv[];
{
	int	i;
	int	so;


/* id socket ? */
	if (argc>1)
		get_id_sock(argv[1], &so);
	else
		get_id_sock("", &so);

/* fermeture de la socket */
	if (close (sock[so])<0)
		ERREUR("close()");

/* mise a jour de la table */
	for (i=so+1; i<nbsock; i++)
		sock[i-1] = sock[i];

	nbsock--;
	if (so==dft_sock)
		dft_sock = -1;
	else if (dft_sock > so)
		dft_sock--;

	return (0);
}


/* Primitive shutdown()
 *=======================================================================
 *
 */

int shutdown_call(argc, argv)
	int	argc;
	char	*argv[];
{
	int	sens = -1;
	static	t_item	list[] = { { "in", 0 }, { "out", 1 }, { "both", 2 } };
	int	so;


/* id socket ? */
	if (argc>1)
		get_id_sock(argv[1], &so);
	else
		get_id_sock("", &so);

/* sens ? */
	if (argc>2)
		get_choice("Sens", list, 3, argv[2], &sens);
	else
		get_choice("Sens", list, 3, "", &sens);

/* fermeture de la socket */
	if (shutdown(sock[so], sens)==-1)
	{
		ERREUR("shutdown()");
		return(-1);
	}
	
	return(0);
}


/* Primitive write()
 *=======================================================================
 *
 */

int 
write_call (argc, argv)
	int             argc;
	char           *argv[];
{
    int	nb;
    /*char	msg[MAX_BUFFER+1]; */
    char * msg;
    int	so;

    msg = (char *) malloc((MAX_BUFFER+1)*sizeof(char));

    /*id socket ? */
    if (argc>1)
        get_id_sock(argv[1], &so);
    else
        get_id_sock("", &so);

    /* message ? */
    /*    nb = MAX_BUFFER; */
    if (argc>2)
        strcpy(msg, argv[2]);
    else
        strcpy(msg, "");

    get_msg(&msg, &nb);


/*      // envoi du message */
    nb = write(sock[so], (void *)msg, (unsigned)strlen(msg));
    if (nb<0)
        {
        ERREUR("write()");
        return(-1);
        }

    free(msg);
    printf("%d octet(s) envoye(s)\n", nb);

    return(0);
}

    
/* Primitive send()
 *=======================================================================
 *
 */

int send_call(argc, argv)
	int	argc;
	char	*argv[];
{
	int	nb;
        char * msg;
	int	total;
	static int	oob;
	static int	dontroute;
	static int	loop;
	static t_flg	flgs[] = {	{ "oob",	&oob },
					{ "dontroute",	&dontroute },
					{ "loop",	&loop } };
	int	so;

        msg = (char *) malloc((MAX_BUFFER+1)*sizeof(char));


/* flags sur la ligne de commande */
	oob = dontroute = loop = 0;
	argc = check_flags(argc, argv, flgs, 3);

/* id socket ? */
	if (argc>1)
		get_id_sock(argv[1], &so);
	else
		get_id_sock("", &so);

/* message ? */
	if (argc>2)
		strcpy(msg, argv[2]);
	else
		strcpy(msg, "");

	get_msg(&msg, &nb);

/* envoi du message */
	total = 0;
	do
	{
		nb = send(sock[so], msg, strlen(msg), 
					MSG_OOB*oob | MSG_DONTROUTE*dontroute);
		if (nb>=0)
			if (loop)
				printf("%d octet(s) envoye(s) (total %d)\n", nb,
								total+=nb);
			else
				printf("%d octet(s) envoye(s)\n", nb);
		else
		{
			ERREUR("send()");
			return(-1);
		}
	}
	while (loop && nb>0);
        free(msg);
	return(0);
}


/* Primitive sendto()
 *=======================================================================
 *
 * Meme chose que la fonction send_call ci-dessus mais en precisant l'hote...
 */

int sendto_call(argc, argv)
	int	argc;
	char	*argv[];
{
	int	nb;
/*  //	char	msg[MAX_BUFFER+1]; */
        char *msg;
	int	total	= 0;
	int	port	= -1;
	struct sockaddr_in sa;
	u_long	addr	= INADDR_NONE;
	static int	oob;
	static int	dontroute;
	static int	loop;
	static t_flg	flgs[] = {	{ "oob",	&oob },
					{ "dontroute",	&dontroute },
					{ "loop",	&loop } };
	
	int	so;

        msg = (char *) malloc((MAX_BUFFER+1)*sizeof(char));

/* flags sur la ligne de commande */
	oob = dontroute = loop = 0;
	argc = check_flags(argc, argv, flgs, 3);

/* id socket ? */
	if (argc>1)
		get_id_sock(argv[1], &so);
	else
		get_id_sock("", &so);

/* host ? */
	if (argc>2)
		get_host(argv[2], &addr);
	else
		get_host("", &addr);

/* port ? */
	if (argc>3)
		get_port(argv[3], &port);
	else
		get_port("", &port);

/* message ? */
/*  	//	nb = MAX_BUFFER; */
	if (argc>4)
		strcpy(msg, argv[4]);
	else
		strcpy(msg, "");

	get_msg(&msg, &nb);

/* initialisation de l'adresse */

	/* bzero((char *)&sa, sizeof(struct sockaddr_in)); cc96-1 */
	memset ((char *) &sa, 0, sizeof (struct sockaddr_in));

	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = addr;
/* P. Sicard attention ici pas de htonl sur l'adresse car ??? */
	sa.sin_port = htons ((u_short) port);

/* envoi du message */
	do
	{
		nb = sendto(sock[so], msg, (unsigned int) strlen(msg),
					MSG_OOB*oob | MSG_DONTROUTE*dontroute,
					(struct sockaddr *)&sa, sizeof(sa));
		if (nb>=0)
			if (loop)
				printf("%d octet(s) envoye(s) (total %d)\n", nb,
								total+=nb);
			else
				printf("%d octet(s) envoye(s)\n", nb);
		else
		{
			ERREUR("sendto()");
			return (-1);
		}
	}
	while (loop && nb>0);
        free(msg);
	return(0);
}


/* Primitive read()
 *=======================================================================
 *
 */

int read_call(argc, argv)
	int	argc;
	char	*argv[];
{
	int	nb	= -1;
	int	lus;
	int	so;
	char	* msg;


/* id socket ? */
	if (argc>1)
		get_id_sock(argv[1], &so);
	else
		get_id_sock("", &so);

/* combien ? */
	nb = 100;
	if (argc>2)
		get_nb("Nb d'octets a lire", argv[2], &nb);
	else
		get_nb("Nb d'octets a lire", "", &nb);
	
	msg  = (char *) malloc((nb+1) * sizeof(char)) ;

/* lecture des donnees */
	lus = read(sock[so], (void *)msg, (unsigned)nb);
	if (lus<0)
	{
		ERREUR("read()");
		return(-1);
	}

	msg[lus] = (char)0;
	printf("%d octet(s) lu(s): message=<%s>\n", lus, msg);
	free(msg);
	return(0);
}


/* Primitive recv()
 *=======================================================================
 *
 */

int recv_call(argc, argv)
	int	argc;
	char	*argv[];
{
	int nb = -1;
	int lus;
	char msg[MAX_BUFFER+1];
	static	int	oob;
	static	int	peek;
	static t_flg	flgs[] = {	{ "oob",	&oob },
					{ "peek",	&peek } };
	int	so;


/* flags sur la ligne de commande */
	oob = peek = 0;
	argc = check_flags(argc, argv, flgs, 2);

/* id socket ? */
	if (argc>1)
		get_id_sock(argv[1], &so);
	else
		get_id_sock("", &so);

/* combien ? */
	nb = 100;
        if (oob)
            nb=1; // Un seul octet transmis en urgent...
        else {
            if (argc>2)
		get_nb("Nb d'octets a lire", argv[2], &nb);
            else
		get_nb("Nb d'octets a lire", "", &nb);
        }
/* lecture des donnees */
	lus = recv(sock[so], msg, nb, MSG_OOB*oob | MSG_PEEK*peek );
	if (lus<0)
	{
		ERREUR("recv()");
		return(-1);
	}

	msg[lus] = (char)0;
	printf("%d octet(s) lu(s): message=<%s>\n", lus, msg);

	return(0);
}


/* Primitive recvfrom()
 *=======================================================================
 *
 */

int recvfrom_call(argc, argv)
	int	argc;
	char	*argv[];
{
	u_long	addr;
	int	port;
	int	nb;
	char	* msg;
	int	lus;
	struct sockaddr_in	sa;
	socklen_t lensa = sizeof(struct sockaddr_in);
	struct hostent *hp;
	static	int	oob;
	static	int	peek;
	static t_flg	flgs[] = {	{ "oob",	&oob },
					{ "peek",	&peek } };
	int	so;


/* flags sur la ligne de commande */
	oob = peek = 0;
	argc = check_flags(argc, argv, flgs, 2);

/* id socket ? */
	if (argc>1)
		get_id_sock(argv[1], &so);
	else
		get_id_sock("", &so);

/* combien ? */
	nb = 100;
        if (oob)
            nb=1; // Un seul octet transmis en urgent...
        else{
            if (argc>2)
		get_nb("Nb d'octets a lire", argv[2], &nb);
            else 
		get_nb("Nb d'octets a lire", "", &nb);
        }	
	msg  = (char *) malloc((nb+1) * sizeof(char)) ;
/* initialisation de l'adresse */

	/* bzero((char *)&sa, sizeof(struct sockaddr_in)); cc96-1 */
	memset ((char *) &sa, 0, sizeof (struct sockaddr_in));

	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = htonl (addr);  /*Modif P.Sicard */
	sa.sin_port = htons ((u_short) port);

/* lecture */
	lus = recvfrom(sock[so], msg, nb, MSG_OOB*oob | MSG_PEEK*peek,
					(struct sockaddr *)&sa, &lensa);
	if (lus<0)
	{
		ERREUR("recvfrom()");
		return(-1);
	}

/* identification de l'origine des donnees */
	hp = gethostbyaddr ((char *) &sa.sin_addr, sizeof (sa.sin_addr), AF_INET);

	if (hp == NULL)
		printf ("Un message de %d octet(s) a ete recu de %s (%d).\n",
			lus, inet_ntoa (sa.sin_addr), ntohs(sa.sin_port));
			/*Modif P. Sicard */
	else
		printf ("Un message de %d octet(s) a ete recu de %s (%d).\n",
			lus, hp->h_name, ntohs(sa.sin_port));
			/*Modif P. Sicard */

	msg[lus] = (char) 0;
	printf ("Message=<%s>\n", msg);
	free(msg);
	return (0);
}
/*_______________________________________________________________________________
 *_______________________________________________________________________________
 */

/* Primitive msocket()
 *=======================================================================
 * permet de creer une socket UDP SANS L'ATTACHER A UN PORT
 *
 */

int msocket_call(argc, argv)
	int	argc;
	char	*argv[];
{
	static	char	*socket_argv[] = { "socket", "udp", 0 };
	int	so;

	so = socket_call(2, socket_argv);
	if (so==-1)
		return(-1);

	printf("Socket UDP no.%d creee...\n", sock[so]);
	/*Modif P. Sicard modification du TTL car par defaut=1 */
	{
	u_char ttl=TTL_MCAST;
	if( setsockopt( sock[so], IPPROTO_IP, IP_MULTICAST_TTL,
		(char *)&ttl, sizeof(ttl) ) == -1 )
             perror( "Attention ttl par defaut  a 1\n" );
	}
	return(0);
}

/*------------------------------------------------------------------------*/

/* Primitive mbind()
 *=======================================================================
 *
 */

int
mbind_call (argc, argv)
	int             argc;
	char           *argv[];
{

/* Declarations locales */

    int so;
    int port;
    struct sockaddr_in sin;

/* "Bind" de la nouvelle socket sur un port */

/* id socket ? */
	if (argc > 1)
		get_id_sock (argv[1], &so);
	else
		get_id_sock ("", &so);

/* port ? */
	if (argc > 3)
		get_port (argv[3], &port);
	else
		get_port ("", &port);

/* Initialisation de la structure necessaire au "bind" */

		memset((char *)&sin, 0, sizeof (sin));
		sin.sin_family = AF_INET;
		sin.sin_port = htons(port);

/* Realisation du "bind" */

		if (bind(sock[so], (struct sockaddr *)&sin, sizeof (sin) ) == -1 )
	          {
		    perror( "bind" );
		    return (-1);
	          }
	return(0);
}

/*------------------------------------------------------------------------*/

/* Primitive mjoin()
 *=======================================================================
 * Permet de rejoindre un groupe multicast
 *
 */

int 
mjoin (argc, argv)
	int             argc;
	char           *argv[];
{

/* Declarations locales */

    int so;
    char line[80];
    char *lineptr;
    struct ip_mreq imr;
    int n = 0;
    unsigned i1, i2, i3, i4, g1, g2, g3, g4;

    u_long addr	= INADDR_NONE;
    char my_host[50];
    
/* id socket ? */
	if (argc > 1)
		get_id_sock (argv[1], &so);
	else
		get_id_sock ("", &so);

/* Quel groupe joindre ? */ 

	if (argc > 2)
	{
		printf("Groupe multicast : %s\n", argv[1]);
		lineptr = argv[2];
	}
	else
	{
		printf("Adresse du groupe multicast ? ");
		fgets( line, 79, stdin );
		lineptr = line;
	}

	while( *lineptr == ' ' || *lineptr == '\t' ) ++lineptr;
	
	if( (n = sscanf( lineptr, "%u.%u.%u.%u", &g1, &g2, &g3, &g4 )) != 4 )
	{
		printf( "bad args\n" );
		return(-1);
	}


/* Sur quelle interface ? */
       /* if (argc>3)
            get_host(argv[3], &addr);
        else
            get_host("", &addr); */
        gethostname(my_host,50);
        get_host(my_host, &addr);
    
/* Initialisation de la structure necessaire au "setsockopt" */

            imr.imr_multiaddr.s_addr = (g1<<24) | (g2<<16) | (g3<<8) | g4;
            imr.imr_multiaddr.s_addr = htonl(imr.imr_multiaddr.s_addr);
            imr.imr_interface.s_addr = addr;
    
/* Realisation du "setsockopt" */

            if( setsockopt( sock[so], IPPROTO_IP, IP_ADD_MEMBERSHIP,
                            (char *)&imr, sizeof(struct ip_mreq) ) == -1 )
                    perror( "can't join group" );
            else printf( "OK : adhesion au groupe %u.%u.%u.%u effectuee...\n", g1, g2, g3, g4);

            return (0);

}

/*------------------------------------------------------------------------*/

/* Primitive mleave()
 *=======================================================================
 * Permet de quitter un groupe multicast
 *
 */

int 
mleave (argc, argv)
	int             argc;
	char           *argv[];
{

/* Declarations locales */

    int so;
    char line[80];
    char *lineptr;
    struct ip_mreq imr;
    int n = 0;
    unsigned i1, i2, i3, i4, g1, g2, g3, g4;

/* id socket ? */
	if (argc > 1)
		get_id_sock (argv[1], &so);
	else
		get_id_sock ("", &so);

/* Quel groupe quitter ? */

	if (argc > 2)
	{
		printf("Groupe multicast : %s\n", argv[1]);
		lineptr = argv[2];
	}
	else
	{
		printf("Adresse du groupe multicast ? ");
		fgets( line, 79, stdin );
		lineptr = line;
	}

	while( *lineptr == ' ' || *lineptr == '\t' ) ++lineptr;
	
	if( (n = sscanf( lineptr, "%u.%u.%u.%u", &g1, &g2, &g3, &g4 )) != 4 )
	{
		printf( "bad args\n" );
		return(-1);
	}


/* Sur quelle interface ? */

	if (argc > 3)
	{
		printf("Adresse locale : %s\n", argv[2]);
		lineptr = argv[3];
	}
	else
	{
		printf("Adresse locale ? ");
		fgets( line, 79, stdin );
		lineptr = line;
	}

	while( *lineptr == ' ' || *lineptr == '\t' ) ++lineptr;

	if( (n = sscanf( lineptr, "%u.%u.%u.%u", &i1, &i2, &i3, &i4 )) != 4 )
	{
		printf( "bad args\n" );
		return(-1);
	}

/* Initialisation de la structure necessaire au "setsockopt" */

		imr.imr_multiaddr.s_addr = (g1<<24) | (g2<<16) | (g3<<8) | g4;
		imr.imr_multiaddr.s_addr = htonl(imr.imr_multiaddr.s_addr);
		imr.imr_interface.s_addr = (i1<<24) | (i2<<16) | (i3<<8) | i4;
		imr.imr_interface.s_addr = htonl(imr.imr_interface.s_addr);

/* Realisation du "setsockopt" */

		if( setsockopt( sock[so], IPPROTO_IP, IP_DROP_MEMBERSHIP,
				(char *)&imr, sizeof(struct ip_mreq) ) == -1 )
		     perror( "can't leave group" );
		else printf( "OK : groupe %u.%u.%u.%u quitte...\n", g1, g2, g3, g4);

	return (0);

}

/*------------------------------------------------------------------------*/

/* Primitive msendto()
 *=======================================================================
 *
 */

int 
msendto_call (argc, argv)
	int             argc;
	char           *argv[];
{

/* Declarations locales */

	int		so;
	struct sockaddr_in sa;
	int             port = 1111;
	int             nb, n;
/*  	char            msg[MAX_BUFFER + 1]; */
        char *msg;
	int             total = 0;
	static int      loop;
	struct in_addr  ifaddr;
	struct sockaddr_in sin;
	char line[80];
	char *lineptr;
	unsigned int g1, g2, g3, g4;
        unsigned int i1, i2, i3, i4 ;

        msg = (char *) malloc((MAX_BUFFER+1)*sizeof(char));

/* Quelle socket utiliser ? */

/* id socket ? */
	if (argc > 1)
		get_id_sock (argv[1], &so);
	else
		get_id_sock ("", &so);

/* A quel groupe envoyer le message ? */

	if (argc > 2)
	{
		printf("Groupe multicast : %s\n", argv[1]);
		lineptr = argv[2];
	}
	else
	{
		printf("Adresse du groupe multicast destinataire ? ");
		fgets( line, 79, stdin );
		lineptr = line;
	}

	while( *lineptr == ' ' || *lineptr == '\t' ) ++lineptr;
	
	if( (n = sscanf( lineptr, "%u.%u.%u.%u", &g1, &g2, &g3, &g4 )) != 4 )
	{
		printf( "bad args\n" );
		return(-1);
	}

/* Sur quelle interface ? */

	if (argc > 3)
	{
		printf("Adresse locale : %s\n", argv[2]);
		lineptr = argv[3];
	}
	else
	{
		printf("Adresse source ? ");
		fgets( line, 79, stdin );
		lineptr = line;
	}

	while( *lineptr == ' ' || *lineptr == '\t' ) ++lineptr;

	if( (n = sscanf( lineptr, "%u.%u.%u.%u", &i1, &i2, &i3, &i4 )) != 4 )
	{
		printf( "bad args\n" );
		return(-1);
	}

/* Socket source deja bindee */
/* modification des options de la socket source */

	ifaddr.s_addr = htonl ((i1<<24) | (i2<<16) | (i3<<8) | i4);

	if (setsockopt (sock[so], IPPROTO_IP, IP_MULTICAST_IF, (char *) &ifaddr.s_addr, sizeof (ifaddr.s_addr)) == -1)
	{
		perror ("can't set multicast source interface");
		return (-1);
	}

/* Numero de port destination ? */
/* modif P. Sicard Port source et port destination pas forcement les memes */
	printf ("Numero de port destination \n");

	if (argc > 4)
		get_port (argv[4], &port);
	else
		get_port ("", &port);


/* Initialisation de l'adresse destination*/

	/* bzero((char *)&sa, sizeof(struct sockaddr_in)); cc96-1 */
	memset ((char *) &sa, 0, sizeof (struct sockaddr_in));

	sa.sin_family = AF_INET;
	sa.sin_port = htons ((u_short) port);
	sa.sin_addr.s_addr = htonl ((g1<<24) | (g2<<16) | (g3<<8) | g4);

/* message ? */
	/*nb = MAX_BUFFER;*/
	if (argc > 4)
		strcpy (msg, argv[4]);
	else
		strcpy (msg, "");

	get_msg (&msg, &nb);

/* envoi du message */
	do
	{
		nb = sendto (sock[so], msg, (unsigned int) strlen (msg),
			     0,
			     (struct sockaddr *) & sa, sizeof (sa));
		if (nb >= 0)
			if (loop)
				printf ("%d octet(s) envoye(s) (total %d)\n", nb,
					total += nb);
			else
				printf ("%d octet(s) envoye(s)\n", nb);
		else
		{
			ERREUR ("sendto()");
			return (-1);
		}
	}
	while (loop && nb > 0);
	free(msg);
	return(0);
}

/*------------------------------------------------------------------------*/

/* Primitive mrecvfrom()
 *=======================================================================
 *
 */

int 
mrecvfrom_call (argc, argv)
	int             argc;
	char           *argv[];
{

/* Declarations locales */

	int count;
	struct timeval start, stop;
	struct sockaddr_in from;
	socklen_t fromlen = 0;
	int read = 0;
	char line[80];
	char *lineptr;
	int so;
	int n;

/* id socket ? */
	if (argc > 1)
		get_id_sock (argv[1], &so);
	else
		get_id_sock ("", &so);

	if (argc > 2)
	{
		printf("Nombre de messages a recevoir : %s\n", argv[2]);
		lineptr = argv[2];
	}
	else
	{
		printf("Combien de messages ? ");
		fgets( line, 79, stdin );
		lineptr = line;
	}

	while( *lineptr == ' ' || *lineptr == '\t' ) ++lineptr;
	
	if( (n = sscanf( lineptr, "%d", &count )) != 1 )
	{
		printf( "bad args\n" );
	}

		gettimeofday(&start, NULL);

		while (count-- > 0) {
		    char buf[BUFSIZ];
		    printf("En attente d'arrivee de messages...\n");
		    fromlen = sizeof(from);
		    read = recvfrom( sock[so], buf, sizeof(buf), 0, (struct sockaddr *)&from, &fromlen );
		    if (read < 0)
		        {
				perror( "recvfrom" );
				return(-1);
			}
			buf[read] = '\0';
			printf ("Un message de %d octet(s) a ete recu.\n",read);
			printf("Message recu : %s\n", buf);
			strcpy(buf, "");

		    fflush(NULL);
	        }

		gettimeofday(&stop, NULL);
		if (fromlen) {
		    printf("Recu de %s en %ld secondes\n",
			   inet_ntoa(from.sin_addr),
			   stop.tv_sec - start.tv_sec);
	        }	
	return(0);
}

/*------------------------------------------------------------------------*/

/* Primitive bsend()
 *=======================================================================
 *
 */

int 
bsend_call (argc, argv)
	int             argc;
	char           *argv[];
{
/* Declarations locales */

	int		so;

/* Quelle socket utiliser ? */

/* id socket ? */
	if (argc > 1)
		get_id_sock (argv[1], &so);
	else
		get_id_sock ("", &so);

/* Realisation du "setsockopt" */

	set_bool_opt(sock[so], 0);

	sendto_call(1, "");

	return(0);
}
